/********************************************************************
 *            dvarchive.c
 *
 *  Tue Aug  1 00:50:52 2006
 *  Copyright  2006  Neil Williams
 *  linux@codehelp.co.uk
 * Copyright (C) 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
 *******************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <ar.h>
#include <archive.h>
#include <archive_entry.h>
#include <fcntl.h>
#include "callbacks.h"
#include "support.h"
#include "dvarchive.h"

#define BLOCKSIZE 10240

typedef enum
{
	DV_OK,
	DV_ERR_NO_FILE,
	DV_ERR_FILE_READ,
	DV_ERR_VERSION,
	DV_ERR_CORRUPT_FILE,
	DV_ERR_NOT_DPKG,
	DV_ERR_COMPRESS,
} DVError;

typedef struct
{
	gint64 size;
	gpointer content;
	gboolean compressed;
}DVFileData;

typedef struct
{
	const gchar *filename;
	gpointer readbuffer;
	gchar *pkg_name;
	gchar *author_name;
	gchar *copyright;
	gchar *control;
	gchar *summary;
	gchar *description;
	gchar *rules;
	gint64 expand_size;
	GHashTable *control_files;
	GHashTable *data_files;
	struct archive *control_tar;
	struct archive *data_tar;
	struct ar_hdr *current;
	gulong skip;
	FILE *pkg;
	GtkWindow *parent;
	DVError error;
} DVContents;

static DVContents *deb = NULL;

enum compression_type
{ CAT, GZ, BZ2 };

void
dv_create (void)
{
	if (!deb)
	{
		deb = g_new0 (DVContents, 1);
		deb->control_files = 
			g_hash_table_new (g_str_hash, g_str_equal);
		deb->data_files = 
			g_hash_table_new (g_str_hash, g_str_equal);
	}
}

void
dv_archive_free (void)
{
	if (!deb)
		return;
	if (deb->control_files)
		g_hash_table_destroy (deb->control_files);
	if (deb->data_files)
		g_hash_table_destroy (deb->data_files);
	if (deb->copyright)
		g_free (deb->copyright);
	if (deb->control)
		g_free (deb->control);
	deb = NULL;
}

void
set_dv_parent (GtkWidget * widget)
{
	if (!deb)
		dv_create();
	deb->parent = GTK_WINDOW (widget);
}

static void
dv_show_error (DVError dve, GtkWidget * widget)
{
	GtkDialog *dialog;
	GtkWindow *parent;
	gboolean fatal;
	gchar *fmt;

	fatal = FALSE;
	fmt = NULL;
	parent = (!widget) ? deb->parent : GTK_WINDOW (widget);
	switch (dve)
	{
		case DV_OK:
			return;
		case DV_ERR_NO_FILE:
		{
			fmt = _("Unable to locate archive file.");
			break;
		}
		case DV_ERR_FILE_READ:
		{
			fmt = _("failed to read archive");
			break;
		}
		case DV_ERR_VERSION:
		{
			fmt = _("unsupported version");
			break;
		}
		case DV_ERR_CORRUPT_FILE :
		{
			fmt = _("possible corrupted file?");
			break;
		}
		case DV_ERR_NOT_DPKG :
		{
			fmt = _("not a debian binary archive");
			break;
		}
		case DV_ERR_COMPRESS :
		{
			fmt = _("unable to decompress file within package.");
			break;
		}
	}
	dialog = GTK_DIALOG (gtk_message_dialog_new (parent,
			GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_ERROR,
			GTK_BUTTONS_CLOSE, _("Error loading file '%s': %s"),
			fmt, deb->filename));
	gtk_dialog_run (dialog);
	gtk_widget_destroy (GTK_WIDGET (dialog));
	if (fatal)
	{
		dv_archive_free ();
		gtk_main_quit ();
	}
}

void
set_deb_file (const gchar * filename)
{
	if (!deb)
		dv_create ();
	deb->filename = filename;
}

void
dv_check_error (void)
{
	dv_show_error (deb->error, NULL);
}

static gint
dv_archive_open (struct archive *a, gpointer client_data)
{	
	DVContents *d = (DVContents*)client_data;

	g_return_val_if_fail (d->readbuffer == NULL, ARCHIVE_FATAL);
	/* deb->pkg already open */
	{
		gchar *size;
		d->skip = 0;
		size = g_strndup (d->current->ar_size, 10);
		d->skip = g_ascii_strtod (size, NULL);
		g_free (size);
		d->readbuffer = malloc (d->skip);
	}
	if( d->readbuffer == NULL ) {
		archive_set_error(a, ENOMEM, _("Out of memory"));
		return ARCHIVE_FATAL;
	}
	return 0;
}

static gint
dv_compressed_open (struct archive *a, gpointer client_data)
{	
	DVContents *d = (DVContents*)client_data;

	g_return_val_if_fail (d->readbuffer == NULL, ARCHIVE_FATAL);
	d->readbuffer = malloc (d->expand_size);
	if( d->readbuffer == NULL ) {
		archive_set_error(a, ENOMEM, _("Out of memory"));
		return ARCHIVE_FATAL;
	}
	return 0;
}

static ssize_t
dv_archive_read (struct archive *a, gpointer client_data,
	gconstpointer *buffer)
{
	ssize_t bytesread;
	DVContents *d = (DVContents*)client_data;

	*buffer = d->readbuffer;
	g_return_val_if_fail (d->readbuffer != NULL, -1);
	if (d->skip == 0)
		return 0;
	bytesread = fread (d->readbuffer, 1, d->skip, d->pkg);
	if (bytesread < 0)
	{
		archive_set_error (a, errno, 
			_("Error reading from file: %m"));
		dv_show_error (DV_ERR_FILE_READ, NULL);
		return -1;
	}
	if (bytesread == 0)
	{
		archive_set_error (a, EIO, 
			_("Unexpected end of file"));
		dv_show_error (DV_ERR_FILE_READ, NULL);
		return -1;
	}
	d->skip -= bytesread;
	return bytesread;
}

static gint
dv_archive_close (struct archive *r, gpointer client_data)
{
	DVContents *d = (DVContents*)client_data;

	g_free(d->readbuffer);
	d->readbuffer = NULL;
	return ARCHIVE_OK;
}

/* from dpkg extract.c */
static gulong
parseheaderlength (const gchar *inh, size_t len)
{
	gchar lintbuf[15];
	gulong r;
	gchar *endp;

	if (memchr (inh, 0, len))
	{
		dv_show_error (DV_ERR_CORRUPT_FILE, NULL);
		return 0;
	}
	g_assert (sizeof (lintbuf) > len);
	memcpy (lintbuf, inh, len);
	lintbuf[len] = ' ';
	*strchr (lintbuf, ' ') = 0;
	r = strtoul (lintbuf, &endp, 10);
	if (*endp)
		dv_show_error (DV_ERR_CORRUPT_FILE, NULL);
	return r;
}

void
prepare_contents (const gchar * filename, GtkWidget * widget)
{
	gint memberlen, error_num;
	gchar *infobuf;
	struct stat stab;
	struct ar_hdr debian_binary;
	struct ar_hdr controlgz;
	struct ar_hdr datagz;
	gchar versionbuf[40];

	if (!deb)
		dv_create ();
	deb->pkg = fopen (filename, "r");
	if (!deb->pkg)
	{
		dv_show_error (DV_ERR_FILE_READ, widget);
		fclose (deb->pkg);
		return;
	}
	if (fstat (fileno (deb->pkg), &stab))
	{
		dv_show_error (DV_ERR_NO_FILE, widget);
		fclose (deb->pkg);
		return;
	}
	if ((!fgets (versionbuf, sizeof (versionbuf), deb->pkg)) ||
		(strcmp (versionbuf, ARMAG)))
	{
		dv_show_error (DV_ERR_VERSION, widget);
		fclose (deb->pkg);
		return;
	}
	fread (&debian_binary, 1, sizeof (struct ar_hdr), deb->pkg);
	if (memcmp (debian_binary.ar_name, "debian-binary   ",
			sizeof (debian_binary.ar_name)) &&
		memcmp (debian_binary.ar_name, "debian-binary/   ",
			sizeof (debian_binary.ar_name)))
		fprintf (stderr, _("file `%.250s' is not a debian "
				"binary archive (try dpkg-split?)"), filename);
	memberlen = parseheaderlength (debian_binary.ar_size,
		sizeof (debian_binary.ar_size));
	infobuf = malloc (memberlen + 1);
	fread (infobuf, 1, memberlen + (memberlen & 1), deb->pkg);
	fread (&controlgz, 1, sizeof (struct ar_hdr), deb->pkg);
	deb->current = &controlgz;
	deb->control_tar = archive_read_new ();
	archive_read_support_compression_gzip (deb->control_tar);
	archive_read_support_format_tar (deb->control_tar);
	error_num = ARCHIVE_OK;
	error_num = archive_read_open (deb->control_tar, deb,
		dv_archive_open, dv_archive_read, dv_archive_close);
	if (error_num != ARCHIVE_OK)
	{
		dv_show_error (DV_ERR_CORRUPT_FILE, widget);
		fclose (deb->pkg);
		return;
	}
	else
	{
		struct archive_entry *entry;
		gchar *locate;
		gchar *content;
		gint size;
		DVFileData *dvfd;

		while (archive_read_next_header (deb->control_tar, 
			&entry) == ARCHIVE_OK)
		{
			size = archive_entry_size (entry);
			locate = g_strdup (archive_entry_pathname (entry));
			content = g_new0(gchar, size);
			archive_read_data (deb->control_tar, content, size);
			dvfd = g_new0 (DVFileData, 1);
			dvfd->size = size;
			dvfd->content = content;
			dvfd->compressed = FALSE;
			g_hash_table_insert (deb->control_files, 
				locate, dvfd);
			if ((0 == strcmp (locate, "./control")) ||
				(0 == strcmp (locate, "control")))
			{
				GtkWidget *display_area;
				GtkTextBuffer *buffer;

				/* todo: pattern match Package: Version: for
				title of window.*/
				deb->control = g_strdup (content);
				display_area = lookup_widget 
					(widget, "display");
				buffer =
					gtk_text_view_get_buffer (GTK_TEXT_VIEW
					(display_area));
				gtk_text_buffer_set_text (buffer, content, -1);
				gtk_text_view_set_buffer (GTK_TEXT_VIEW 
					(display_area),	buffer);
			}
			archive_read_data_skip (deb->control_tar);
		}
	}
	dv_archive_close (deb->control_tar, deb);
	{
		/* if the size of control.tar.gz is odd,
		seek one byte to the start of data.tar.gz */
		gulong bytes_left;
		off_t s;
		gint seek;
		bytes_left = strtoul(controlgz.ar_size,NULL,10);
		seek = (bytes_left & 1);
		s = fseek(deb->pkg,seek,SEEK_CUR);
		if( s == (off_t)-1 ) {
			dv_show_error (DV_ERR_FILE_READ, widget);
			fclose (deb->pkg);
			return;
		}
	}
	fread (&datagz, 1, sizeof (struct ar_hdr), deb->pkg);
	deb->current = &datagz;
	deb->data_tar = archive_read_new ();
	archive_read_support_compression_gzip (deb->data_tar);
	archive_read_support_format_tar (deb->data_tar);
	error_num = ARCHIVE_OK;
	error_num = archive_read_open (deb->data_tar, deb,
		dv_archive_open, dv_archive_read, dv_archive_close);
	if (error_num != ARCHIVE_OK)
		dv_show_error (DV_ERR_CORRUPT_FILE, widget);
	else
	{
		struct archive_entry *entry;
		gpointer *content;
		gint size;
		gchar *locate;
		DVFileData *dvfd;

		while (archive_read_next_header (deb->data_tar, &entry) ==
			ARCHIVE_OK)
		{
			locate = g_strdup (archive_entry_pathname (entry));
			size = archive_entry_size (entry);
			content = g_new0 (gpointer, size);
			archive_read_data (deb->data_tar, content, size);
			if (!strcmp (locate, "./rules"))
				deb->rules = g_strdup ((gchar*)content);
			dvfd = g_new0 (DVFileData, 1);
			dvfd->size = size;
			dvfd->content = content;
			dvfd->compressed = g_str_has_suffix (locate, ".gz");
			g_hash_table_insert (deb->data_files, locate, dvfd);
			archive_read_data_skip (deb->data_tar);
		}
		archive_read_finish (deb->data_tar);
	}
	fclose (deb->pkg);
}

static void
add_control_files (gpointer key, gpointer value, gpointer user_data)
{
	DVIter *dvi = (DVIter *) user_data;
	gchar *path = (gchar *) key;
	DVFileData *dvfd = (DVFileData*)value;
	gchar *size_str;

	if (g_str_has_suffix(path, "/"))
		return;
	size_str = g_strdup_printf ("%" G_GINT64_FORMAT, dvfd->size);
	gtk_tree_store_append (dvi->store, dvi->child_iter, 
		dvi->parent_iter);
	gtk_tree_store_set (dvi->store, dvi->child_iter, PKG_COLUMN,
		g_filename_to_utf8 ((gchar*)key, -1, NULL, NULL, NULL),
		SIZE_COLUMN, size_str, LOCATION_COLUMN,
		g_filename_to_utf8 ((gchar*)key, -1, NULL, NULL, NULL), -1);
}

static void
add_data_files (gpointer key, gpointer value, gpointer user_data)
{
	DVIter *dvi = (DVIter *) user_data;
	gchar *path = (gchar *) key;
	gchar *file = g_path_get_basename (path);
	DVFileData *dvfd = (DVFileData*)value;
	gchar *size_str;

	if (g_str_has_suffix(path, "/"))
		return;
	size_str = g_strdup_printf ("%" G_GINT64_FORMAT, dvfd->size);
	gtk_tree_store_append (dvi->store, dvi->child_iter, 
		dvi->parent_iter);
	gtk_tree_store_set (dvi->store, dvi->child_iter, PKG_COLUMN,
		g_filename_to_utf8 (file, -1, NULL, NULL, NULL),
		SIZE_COLUMN, size_str, LOCATION_COLUMN,
		g_filename_to_utf8 (path, -1, NULL, NULL, NULL), -1);
}

void
dv_archive_get_control (DVIter * dvi)
{
	g_hash_table_foreach (deb->control_files, add_control_files, dvi);
	/* add package, etc. from DVContents */
}

void
dv_archive_get_data (DVIter * dvi)
{
	g_hash_table_foreach (deb->data_files, add_data_files, dvi);
}

static gboolean store_set = FALSE;

static void
columns_reset (gpointer value, gpointer user_data)
{
	GtkTreeView *treeview;
	GtkTreeViewColumn *column;
	column = (GtkTreeViewColumn*)value;
	treeview = (GtkTreeView*)user_data;
	gtk_tree_view_column_clear (column);
	gtk_tree_view_remove_column (treeview, column);
}

void
open_deb (const gchar * filename, GtkWidget * widget)
{
	GtkWidget *treeview;
	GtkTreeStore *store;
	GtkTreeModel *model;
	GtkTreeIter parent_iter, orig_iter;
	GtkTreeIter deb_iter, tar_iter, file_iter;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	DVIter dvi;

	if (!filename)
		return;
	prepare_contents (filename, widget);
	treeview = lookup_widget (widget, "treeview");
	model = gtk_tree_view_get_model (GTK_TREE_VIEW(treeview));
	if (store_set)
	{
		GList *columns = NULL;
		store = GTK_TREE_STORE (model);
		gtk_tree_store_clear (store);
		columns = gtk_tree_view_get_columns 
			(GTK_TREE_VIEW(treeview));
		g_list_foreach (columns, columns_reset, treeview);
	}
	else
	{
		store = gtk_tree_store_new (N_COLUMNS,
			G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
		gtk_tree_view_set_model (GTK_TREE_VIEW (treeview),
			GTK_TREE_MODEL (store));
		store_set = TRUE;
	}
	gtk_tree_store_append (store, &parent_iter, NULL);
	gtk_tree_store_set (store, &parent_iter, PKG_COLUMN, "name",
		SIZE_COLUMN, "size", LOCATION_COLUMN, filename, -1);
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes
		(_("Package"), renderer, "text", PKG_COLUMN, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
	column = gtk_tree_view_column_new_with_attributes
		(_("Size"), renderer, "text", SIZE_COLUMN, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
	column = gtk_tree_view_column_new_with_attributes
		(_("Location"), renderer, "text", LOCATION_COLUMN, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
	gtk_tree_store_append (store, &deb_iter, &parent_iter);
	gtk_tree_store_set (store, &deb_iter, PKG_COLUMN,
		_("Debian data"), SIZE_COLUMN, NULL, LOCATION_COLUMN, 
		NULL, -1);
	dvi.store = store;
	dvi.child_iter = &file_iter;
	dvi.parent_iter = &deb_iter;
	dv_archive_get_control (&dvi);

	gtk_tree_store_append (store, &orig_iter, &parent_iter);
	gtk_tree_store_set (store, &orig_iter, PKG_COLUMN,
		_("Upstream data"), SIZE_COLUMN, "", LOCATION_COLUMN, "", -1);
	dvi.store = store;
	dvi.child_iter = &tar_iter;
	dvi.parent_iter = &orig_iter;
	dv_archive_get_data (&dvi);
	gtk_tree_view_expand_all (GTK_TREE_VIEW(treeview));
	gtk_widget_show_all (treeview);
}

void
dv_archive_preload (GtkWidget * widget)
{
	if (deb)
		open_deb (deb->filename, GTK_WIDGET (widget));
}

gchar *
dv_get_control_content (const gchar * file)
{
	DVFileData *dvfd;

	dvfd = (DVFileData*)
		g_hash_table_lookup (deb->control_files, file);
	if (!dvfd)
		return NULL;
	return (g_utf8_validate (dvfd->content, -1, NULL)) ? 
		dvfd->content : NULL;
}	

gchar *
dv_get_data_content (const gchar * file)
{
	gchar *check;
	DVFileData *dvfd;
	dvfd = (DVFileData*)
		g_hash_table_lookup (deb->data_files, file);
	if (!dvfd)
		return NULL;
	if (dvfd->compressed)
	{
		struct archive *expand;
		struct archive_entry *entry;
		gint error_num;
		gchar *content;
		gint size;

		check = NULL;
		expand = archive_read_new ();
		archive_read_support_compression_gzip (expand);
		error_num = ARCHIVE_OK;
		error_num = archive_read_open (expand, deb,
			dv_compressed_open, dv_archive_read, 
			dv_archive_close);
		if (error_num != ARCHIVE_OK)
		{
			fprintf (stderr, _("open data.tar.gz within "
				"'%s' failed: %d:%d:%s\n"),
				deb->filename,
			error_num, archive_errno (expand),
			archive_error_string (expand));

			dv_show_error (DV_ERR_COMPRESS, NULL);
			return NULL;
		}
		if (archive_read_next_header (expand, &entry) ==
			ARCHIVE_OK)
		{
			size = archive_entry_size (entry);
			content = g_new0 (gchar, size);
			archive_read_data (deb->data_tar, content, size);
		}
		else
		{
			dv_show_error (DV_ERR_COMPRESS, NULL);
			return NULL;
		}
		check = g_strndup (content, size);
		archive_read_finish (expand);
		g_free (content);
		if (!check)
			return NULL;
	}
	else
	{
		check = dvfd->content;
	}
	return (g_utf8_validate (check, -1, NULL)) ? check : NULL;
}
